@@ -4,7 +4,7 @@ use strict;
 use FileHandle;
 
 use vars qw( $VERSION );
-$VERSION = '0.06';
+$VERSION = '0.07';
 
 =head1 NAME
 
@@ -13,8 +13,12 @@ Audio::Wav::Read - Module for reading Microsoft WAV files.
 =head1 SYNOPSIS
 
     use Audio::Wav;
+
     my $wav = new Audio::Wav;
     my $read = $wav -> read( 'filename.wav' );
+OR
+    my $read = Audio::Wav -> read( 'filename.wav' );
+
     my $details = $read -> details();
 
 =head1 DESCRIPTION
@@ -44,21 +48,41 @@ sub new {
     my $handle = new FileHandle "<$file";
 
     my $self = {
-	'real_size'	=> $size,
-	'file'		=> $file,
-	'handle'	=> $handle,
-	'tools'		=> $tools,
+        'real_size' => $size,
+        'file'      => $file,
+        'handle'    => $handle,
+        'tools'     => $tools,
     };
 
-    bless $self, $class;
+    bless $self, $class; 
 
     unless ( defined $handle ) {
-	$self -> _error( "unable to open file ($!)" );
-	return $self;
+        $self -> _error( "unable to open file ($!)" );
+        return $self;
     }
 
     binmode $handle;
 
+BEGIN {
+    eval { require Inline::C };
+    if($@) {
+        $Audio::Wav::Read::_has_inline = 0;
+    } else {
+        $Audio::Wav::Read::_has_inline = 1;
+    }
+}
+
+    if( $Audio::Wav::Read::_has_inline ) {
+        local $/ = undef;
+        my $c_string = <DATA>; 
+        Inline->import(C => $c_string);
+    } else {
+        #TODO: do we have a reference to $tools here if using shortcuts?
+        if( $tools && $tools -> is_debug() ) {
+            warn "can't load Inline, using slow pure perl reads\n";
+        }
+    }
+
     $self -> {'data'} = $self -> _read_file();
     my $details = $self -> details();
     $self -> _init_read_sub();
@@ -223,42 +247,55 @@ Each element is a channel i.e ( left, right ).
 The numbers will be in the range;
 
     where $samp_max = ( 2 ** bits_per_sample ) / 2
-    -$samp_max to +$samp_max
+    -$samp_max to +$samp_max 
 
 =cut
 
-sub read {
-    my $self = shift;
-    my $val;
-    my $block = $self -> {'data'} -> {'block_align'};
-    return () if $self -> {'pos'} + $block > $self -> {'data'} -> {'data_finish'};
-    $self -> {'pos'} += read( $self -> {'handle'}, $val, $block );
-    return () unless defined( $val );
-    return &{ $self -> {'read_sub'} }( $val );
-}
+# read is generated by _init_read_sub
+sub read { die "call _init_read_sub first"; };
 
 sub _init_read_sub {
     my $self = shift;
-    my $details = $self -> {'data'};
-    my $channels = $details -> {'channels'};
-    my $sub;
+    my $details        = $self -> {'data'};
+    my $channels       = $details -> {'channels'};
+    my $block          = $details -> {'block_align'};
+    my $handle         = $self -> {'handle'};
+    my ($read_op, $read_op_c);
     if ( $details -> {'bits_sample'} <= 8 ) {
-	my $offset = ( 2 ** $details -> {'bits_sample'} ) / 2;
-	$sub = sub { return map $_ - $offset, unpack( 'C'.$channels, shift() ) };
+        my $offset = ( 2 ** $details -> {'bits_sample'} ) / 2;
+        $read_op = q{ return map $_ - $offset, unpack( 'C'.$channels, shift() ) };
     } else {
-	if ( $self -> {'tools'} -> is_big_endian() ) {
-	    $sub = sub {
-		return unpack( 's' . $channels,			# 3. unpack native as signed short
-		    pack( 'S' . $channels,			# 2. pack native unsigned short
-			unpack( 'v' . $channels, shift() )	# 1. unpack little-endian unsigned short
-		    )
-		);
-	   };
-	} else {
-	    $sub = sub { return unpack( 's' . $channels, shift() ) };
-	}
+        if ( $self -> {'tools'} -> is_big_endian() ) {
+            $read_op = q{
+                return unpack( 's' . $channels,             # 3. unpack native as signed short
+                    pack( 'S' . $channels,                  # 2. pack native unsigned short
+                        unpack( 'v' . $channels, shift() )  # 1. unpack little-endian unsigned short
+                    )
+                );
+            };
+        } else {
+            $read_op = q{ return unpack( "s" . $channels, $val ) };
+        }
+    }
+    $self -> {'read_sub_string'} = q[
+        sub {
+            my $val;
+### unnecessary - read() deals with this case already
+#           return () if $self -> {'pos'} + $block > $self -> {'data'} -> {'data_finish'};
+            $self -> {'pos'} += read( $handle, $val, $block );
+            return () unless defined( $val );
+            ] . $read_op . q[
+        };
+    ];
+    if( $Audio::Wav::Read::_has_inline ) {
+        init( $handle, $details->{'bits_sample'}/8, $channels,
+            $self -> {'tools'} -> is_big_endian() ? 1 : 0);
+        *read = \&read_c;
+    } else {
+        my $read_sub = eval $self -> {'read_sub_string'};
+        $self -> {'read_sub'} = $read_sub; #in case any legacy code peaked at that
+        *read = \&$read_sub;  
     }
-    $self -> {'read_sub'} = $sub;
 }
 
 =head2 position
@@ -693,3 +730,108 @@ sub _error {
 =cut
 
 1;
+
+__DATA__
+
+//NOTE: 16, 32 bit audio do *NOT* work on big-endian platforms yet!
+//verified formats (output is identical output to pureperl):
+// 1 channel signed   16 little endian
+// 2 channel signed   16 little endian
+// 1 channel unsigned  8 little endian
+// 2 channel unsigned  8 little endian
+//verified "looks right" on these formats:
+// 1 channel signed   32 little endian
+// 2 channel signed   32 little endian
+// 1 channel signed   24 little endian
+// 2 channel signed   24 little endian
+
+//maximum number of channels per audio stream
+#define MAX_CHANNELS 10
+//maximum number of bytes per sample (in one channel)
+#define MAX_SAMPLE 4
+
+FILE *handle;
+int sample_size;
+int channels;
+int big_end;
+int is_signed;
+char buf[MAX_SAMPLE];
+SV* retvals[MAX_CHANNELS];
+
+void init(FILE *fh, int ss, int ch, int be) {
+    int i;
+    handle = fh;
+    sample_size = ss;
+    channels = ch;
+    big_end = be;
+    is_signed = (ss != 1); //TODO: is this really right?
+    for(i=0; i<MAX_CHANNELS; i++) {
+        retvals[i] = newSV(0);
+    }
+}
+
+void read_c(void *self) {
+    int samples[MAX_CHANNELS];
+    int nread;
+    Inline_Stack_Vars;
+    Inline_Stack_Reset;
+    int i, s;
+    for(i=0; i<channels; i++) {
+        // having fread in the loop is probably slightly less efficient,
+        // but it avoids byte alignment problems and fread is buffered,
+        // so it "shouldn't be a problem" (tm). more info:
+        // http://www.eventhelix.com/RealtimeMantra/ByteAlignmentAndOrdering.htm
+        nread = fread( buf, sample_size, 1, handle );
+        if( !nread ) {
+            if( feof( handle ) && i ) {
+                perror("got EOF mid-sample!");
+            } else if( ferror( handle ) ) {
+                perror("io error");
+            }
+            break;
+        }
+        switch(sample_size) {
+            case 4:
+                if(big_end) {
+                    s = buf[0]; buf[0] = buf[3]; buf[3] = s;
+                    s = buf[1]; buf[1] = buf[2]; buf[2] = s;
+                }
+                s = is_signed ?
+                    *((int32_t *)buf) :
+                    *((uint32_t *)buf)-2147483648;
+                break;
+            case 3:
+                //TODO: test this!
+                if(big_end) { s = buf[0]; buf[0] = buf[2]; buf[2] = s; }
+                s = *((uint32_t *)buf);
+                if(big_end) { s = (s & 0xffffff00) >> 8; }
+                else        { s = s & 0x00ffffff; }
+                //make negative via 2s compliment if data is signed
+                //and the sign bit is set
+                if ( is_signed ) {
+                    if ( s & 0x00800000 ) {
+                        s = -((~s & 0x00ffffff)+1);
+                    }
+                } else {
+                    //we *always* return signed data
+                    s -= 8388608;
+                }
+                break;
+            case 2: 
+                if(big_end) { s = buf[0]; buf[0] = buf[1]; buf[1] = s; }
+                s = is_signed ?
+                    *((int16_t *)buf) :
+                    *((uint16_t *)buf)-32768;
+                break;
+            case 1:
+                //note: Audio::Wav *always* returns signed data
+                s = is_signed ?
+                    *((int8_t *)buf) :
+                    *((uint8_t *)buf)-128;
+                break;
+        }
+        sv_setiv(retvals[i], s);
+        Inline_Stack_Push(retvals[i]);
+    }
+    Inline_Stack_Done;
+}
